/* algnum.gp */

/*close(x,y,tol=1e-10) = abs(x-y) <= tol*(1+max(abs(x), abs(y)))*/
close(x,y,tol=1e-10) = abs(x-y) < tol

/*Cohen p. 158 tells us to do exactly what we're doing:
 Algebraic numbers are represented as a minimal polynomial in x and a
   floating point approximation */
u.minpol = u[1]
u.approx = u[2]

aninit(mp, ap, findroot=1) = {
  local(r, a=0);
  
  if(findroot,
     r = polroots(mp);
     for(i=1, #r, if(abs(r[i]-ap)<abs(a-ap), a=r[i]));
     ap = a);
  
  if(!close(subst(mp,x,ap), 0),
     error("aninit: bad approximation ", [mp,ap]));
     
  [mp,ap]
}

isalgnum(obj) = type(obj)=="t_VEC" && #obj==2 && type(obj[1])=="t_POL"
                
addhelp(aninit,	"aninit(p,t,{f=1}): algebraic number with minimal polynomial p"\
	"(the variable is x) and floating point approximation t; if f=1")
addhelp(minpol, "a.minpol: minimal polynomial of algebraic number a")
addhelp(approx, "a.approx: floating point approximation of algebraic number a")
addhelp(isalgnum, "isalgnum(x): 1 iff x is an algebraic number")

andegree(u) = poldegree(u.minpol)
addhelp(andegree, "andegree(u): degree of the algebraic number u")

ancompositum(u,v,return_uv = 0) =
{
  local(c = polcompositum(u.minpol, v.minpol, 1), p, w);
  for(i=1, #c,
    p = v.approx + c[i][4] * u.approx;
    if(close(u.approx, subst(lift(c[i][2]), x, p)) &&
       close(v.approx, subst(lift(c[i][3]), x, p)),
       w = aninit(c[i][1], p);
       return(if(return_uv,[w, c[i][2], c[i][3]], w))));
  error("compositum not found for " u " and " v)
}
addhelp(ancompositum, "ancompositum(u,v,{return_uv = 0}): return a primitive "\
        " element of Q(u,v), along with polmods giving u and v in terms of it "\
        " if optional parameter return_uv = 1")

anrat(r) =
{
  local(t = type(r)); 
  if(t!="t_FRAC" && t!="t_INT", error("anrat requires a rational number"));
  aninit(x-r, r)
}
addhelp(anrat, "anrat(r): consider a rational number r as an algebraic number")

anadd(u,v) =
{
  local(c = ancompositum(u,v,1));
  aninit(minpoly(c[2] + c[3]), u.approx + v.approx)
}
addhelp(anadd, "anadd(u,v): sum of the algebraic numbers u and v")

/* Takes as input an alegbraic number u. Gives as output the negative of u.*/
anneg(u) = aninit(subst(u.minpol,x, -x), -u.approx)
addhelp(anneg, "anneg(u): negative of the algebraic number u.")

anmul(u,v) =
{
  local(c = ancompositum(u,v,1));
  aninit(minpoly(c[2] * c[3]), u.approx * v.approx)
}
addhelp(anmul, "anmul(u,v): product of the algebraic numbers u and v")
     
ansum(w) =
{
  local(v = anvecinit(w));
  ansubst(sum(i=1, #w, v.data[i]), x, v.prim)
}
addhelp(ansum, "ansum(w): sum of the algebraic numbers in vector w")

anprod(w) =
{
  /*
  local(v = anvecinit(w));
  ansubst(prod(i=1, #w, v.data[i]), x, v.prim)
  */
  local(p = anrat(1));
  for(i=1, #w, p = anmul(p, w[i]);if(debug,print(p)));
  p
}
addhelp(anprod, "anprod(w): product of the algebraic numbers in vector w")

ansqrt(u) =
{
  local(v = u, p = v.minpol, r = sqrt(v.approx));
  aninit(irrfactor(subst(p, x, x^2), r), r)
}
addhelp(ansqrt, "ansqrt(u): square root of the algebraic number u")

aneq(u,v) = (u.minpol == v.minpol) && close(u.approx, v.approx)
addhelp(aneq, "aneq(u,v): 1 iff the algebraic numbers u and v are equal")
     
modulus(p) = component(p,1)

ansubst(p,v,u) =
{
  if(type(p)=="t_POLMOD",
     if(modulus(p) != u.minpol,
        error("anpol: modulus of polmod not equal to the minimal polynomial of"\
              " the algebraic number")),
     p = Mod(p, u.minpol));
  aninit(minpoly(p), subst(lift(p), v, u.approx))
}
addhelp(ansubst, "ansubst(p,x,u): substitute the algebraic number u for x in p")

irrfactor(p,r) =
{
  local(f = factor(p));
  for(i=1, matsize(f)[1], if(close(subst(f[i,1], x, r), 0), return(f[i,1])));
  error("no factor of " p " vanishes at " r)
}
addhelp(irrfactor, "irrfactor(p,r): irreducible factor of p that"\
	" vanishes at r")
        

/* Takes a positive integer n as input; gives as output the minimal
polynomial of 2*cos(Pi/n), expressed in terms of the variable x.*/
/* Does this by equating the lengths of two diagonals of the regular
   n-gon, both of which are expressed in terms of 2*cos(Pi/n).*/
an2cos(n) = {
  local(g=divisors(n), h=vector(#g), k=1);

  if(n==1, return(anrat(-2)));
  if(n%2,
    for(X=1, #g, h[X]=ngondiag((g[X]+1)/2)-ngondiag((g[X]-1)/2)),
    for(X=1, #g, h[X]=ngondiag(g[X])));
  for(X=1, #g, k*=h[X]^(moebius(n/g[X])));
  aninit(k, 2.0 * cos(Pi/n))
}
addhelp(an2cos, "an2cos(n): the minimal polynomial of 2*cos(Pi/n), "\
        "expressed in terms of the variable x.")

anminus2cos(n) = anmul(anrat(-1),an2cos(n))

/* Computes the length of the k-th diagonal of the regular n-gon
   with side length 1. Expresses this length in terms of x = 2*cos(Pi/n).
   By default, the 0-th diagonal has length 0, and the 1st has length 1.*/
/* Uses a recursive computation derived from similar triangles
   (or Ptolemy's theorem.)*/

ngondiag(k) = {
  if(k==0, 0,
     if(k==1, 1,
        x*ngondiag(k-1)-ngondiag(k-2)))
}
addhelp(ngondiag, "ngondiag(k): the length of the k-th diagonal of the "\
        "regular n-gon with side length 1. Expresses this length in terms "\
        "of x = 2*cos(Pi/n). By default, the 0-th diagonal has length 0, and "\
        "the 1st has length 1.")
