/* guess_and_check_gram.gp */

/* This is a sequence of commands that allow the user to increase the precision
of input data consiting of face_normals to a polyhedron, guess algebraic numbers
from high precision decimal values for these face_normals and then to check that
these values are correct by constructing the gram matrix, and checking that the
(i,j)th entry is -2cos(Pi/n_ij) where n_ij is the label on the ij-th edge (or 1
if i=j).*/

guess_minpoly(approx,degree = 30) = irrfactor(algdep(approx,degree),approx)


guess_exact_normals(face_normals,degree_to_guess = 30) =
{
  matrix(matsize(face_normals)[1],4,i,j,
         if(verbose, print("guessing ", [i,j]));
         aninit(guess_minpoly(face_normals[i,j],degree_to_guess),
                face_normals[i,j]))
}

guess_exact_gram_matrix(faces,degree_to_guess = 30) =
{
  local(size_m = matsize(faces)[1],
        quadform = matdiagonal([-2,2,2,2]),
        gram = faces*quadform*faces~);

  matrix(size_m,size_m,i,j,
         if(verbose, print("guessing ", [i,j]));
         aninit(guess_minpoly(gram[i,j],degree_to_guess), gram[i,j]));
}

anmatmul_old(A,B) =
{
  local(p = anrat(1), sizeA = matsize(A), sizeB=matsize(B), C);
  
  C = matrix(sizeA[1],sizeB[2]);
  for (i=1,sizeA[1],
    for (j=1,sizeB[2],
      if(verbose, print([i,j]));
      Vector = vector(sizeA[2],k,anmul(A[i,k],B[k,j]));
      C[i,j] = ansum(Vector)));
  C
}

check_guessed_normals(guessed_normals, edge_labels) =
{
  local(size = matsize(guessed_normals)[1],
        gram = matrix(size,size,i,j,anrat(1)),
        quadform = matrix(4,4,i,j,anrat(0)),
        correct);

  quadform[1,1] = anrat(-2);
  quadform[2,2] = anrat(2);
  quadform[3,3] = anrat(2);
  quadform[4,4] = anrat(2);

  gram = anmatmul_old(guessed_normals,quadform);
  if(verbose, print("multiplication 1"));
  gram = anmatmul_old(gram,guessed_normals~);
  if(verbose, print("multiplication 2"));

  correct = 1;
  for(i=1,size,
    for(j=1,size,
      if(verbose, print([i,j]));
      if(edge_labels[i,j] != 0, 
        if(!aneq(gram[i,j], anminus2cos(edge_labels[i,j])),
          correct = 0;
          if(verbose, print("failed: ", [i,j]))))));

  if (correct, return(gram));
  error("the normal vectors that you entered fail to saitisfy "\
        "the dihedral relations");
}

poly_function (B,edge_labels,face_i,face_j,face_k) =
{
  local(N=matsize(edge_labels)[1],
        A = matrix(4*N,1),
        Vectors = matrix(N,4));

  for(i=1,N,
    for(j=1,4,
      Vectors[i,j] = B[4*(i-1)+j,1]));

  A[1,1] = Vectors[face_i,1];
  A[2,1] = Vectors[face_i,2];
  A[3,1] = Vectors[face_i,3];
  A[4,1] = Vectors[face_j,1];
  A[5,1] = Vectors[face_j,2];
  A[6,1] = Vectors[face_k,1];

  count = 7;

  for(i=1,N,
    for(j=i,N,
      if (edge_labels[i,j] > 0,
        A[count,1] = -Vectors[i,1]*Vectors[j,1] + Vectors[i,2]*Vectors[j,2] +\
                      Vectors[i,3]*Vectors[j,3] + Vectors[i,4]*Vectors[j,4] +\
                      cos(Pi/edge_labels[i,j]);
      count = count + 1)));
  A
}

newton_step(V,edge_labels,face_i,face_j,face_k) = 
{
  local(N = matsize(edge_labels)[1],
        B = matrix(4*N,1), C, D,
        J = matrix(4*N,4*N),
        count, fval);

  for(i=1,N,
    for(j=1,4,
       B[4*(i-1)+j,1] = V[i,j]));

  J[1,4*(face_i-1)+1] = 1;
  J[2,4*(face_i-1)+2] = 1;
  J[3,4*(face_i-1)+3] = 1;
  J[4,4*(face_j-1)+1] = 1;
  J[5,4*(face_j-1)+2] = 1;
  J[6,4*(face_k-1)+1] = 1;
  
  
  count = 7;
  
  for (i=1,N,
    for (j=i,N,
      if(edge_labels[i,j] > 0,
        if (i==j,
          J[count,4*(i-1)+1] = -2*V[i,1];
          J[count,4*(i-1)+2] = 2*V[i,2];
          J[count,4*(i-1)+3] = 2*V[i,3];
          J[count,4*(i-1)+4] = 2*V[i,4];
          count = count +1;
          ,
          J[count,4*(i-1)+1] = -V[j,1];
          J[count,4*(i-1)+2] = V[j,2];
          J[count,4*(i-1)+3] = V[j,3];
          J[count,4*(i-1)+4] = V[j,4];
          J[count,4*(j-1)+1] = -V[i,1];
          J[count,4*(j-1)+2] = V[i,2];
          J[count,4*(j-1)+3] = V[i,3];
          J[count,4*(j-1)+4] = V[i,4];
          count = count +1))));

  C = poly_function(B,edge_labels,face_i,face_j,face_k);
  D = B - matsolve(J,C);

  for(i=1,N,
    for(j=1,4,
      V[i,j] = D[4*(i-1)+j,1]));

  fval = norml2(C);

  return([V,fval]);
}

newton_iter(face_normals,edge_labels,tol) = 
{
  local(A, accuracy, normals);
  A = newton_step(face_normals,edge_labels,1,2,3);
  normals = A[1];
  accuracy = A[2];
  while(accuracy > tol,
    A = newton_step(normals, edge_labels,1,2,3);
    normals = A[1];
    accuracy = A[2]);
  normals
}


