/* inv_tr_field.gp */

invariant_trace_field(g,bfs = 1) = ancompositum(littlek(g), reducefield(ansqrt(Mdisc(g,bfs)))) 
/* As defined in Machlachlan and Reid.  A commensurability invariant.*/
/*Reduce the field generators at inbetween step.*/


/* Takes as input a matrix m of algebraic numbers and a vector path of indices. Gives as output the product of all entries of m with coordinates of the form [path[i], path[i+1]].*/
/* Multiplication uses the function anmul.*/

pathprod(m, path) = anprod(vector(#path-1,i, m[path[i],path[i+1]]))

addhelp(pathprod, "pathprod(m, path): the product of all entries of m with coordinates of the form [path[i], path[i+1]].")

/* Takes as input a matrix m of algebraic numbers and a vector fourpaths of
vectors each of whose entries are indices giving a path in m. Gives as output a
square matrix of size length(fourpaths) whose entry at position [i,j] is the
inner product of the two path product vectors coming from fourpaths.*/
/* Multiplication uses the function anmul.*/

matpaths(m, fourpaths) = 
{
  matrix(#fourpaths,#fourpaths,i,j,
    anmul(m[fourpaths[i][#fourpaths[i]],fourpaths[j][#fourpaths[j]]],
      anmul(pathprod(m,fourpaths[i]),pathprod(m,fourpaths[j]))))
}
addhelp(matpaths, "matpaths(m, paths): matrix of inner products of vectors "\
        "$v_{i_1, \ldots, i_r}$, where the sequences $i_1, \ldots, i_r$ "\
        "represent paths in the matrix m.")

/* Takes as input a square matrix m and two indices k and l, which must be
between 1 and the size of m. Gives as output the matrix m with row k and column
l removed.*/

matcofactor(m, k, l) = {
  matrix(matsize(m)[1]-1,matsize(m)[2]-1, i,j, 
    m[i+(i>=k),j+(j>=l)])
}
addhelp(matcofactor, "matcofactor(m, i, j): matrix m with row i and column j removed.")

/* Takes as input a matrix of algebraic numbers. Gives as output the determinant
of the matrix, expressed as an algebraic number.*/

matdetan(m) = {
  local(v);
  if(matsize(m)==[1,1],
    m[1,1],
    v = vector(length(m), i,
               anmul(if(i%2,m[1,i],anneg(m[1,i])),
                     matdetan(matcofactor(m,1,i))));
    ansum(v))
}
addhelp(matdetan, "matdetan(m): determinant of the matrix m of algebraic numbers.")

/* Calculates the discriminant of the space (M, q), given a gram matrix m in the
proper form. The proper form means that the first three faces have linearly
independent normals, and the first face is adjacent to the second and the
third.*/
Mdisc(m,bsf = 1) = matdetan(Mmatrix(m,bfs))
addhelp(Mdisc, "Mdisc(m): the discriminant of the space (M, q) derived from "\
               "the gram matrix m.")

Mmatrix(m, bfs=1) = {
  local(paths = if(bfs, pathlist(m), otherpathlist(m)), d = anrat(0)); 
  
  for(i=4, #paths,
    four = [paths[1], paths[2], paths[3], paths[i]];
    d = matdetan(matpaths(m, four));
    if(!aneq(d,anrat(0)), return(matpaths(m,four))));
  error("path products of "m" must be linearly independent.")
}
addhelp(Mmatrix, "Mmatrix(G): given a Gram Matrix G, returns the 4x4 matrix called M by Maclachlan and Reid providing the quadratic form spanned by the vectors v...")

/* Returns 1 if a is an element of the vector v; returns 0 otherwise.*/

iselement(a, v) = {
  for(i=1, #v,
    if(a==v[i], return(1)));
  0
}
addhelp(iselement, "iselement(a,v): true if and only if a is an element of the vector v.")

/* Does a breadth-first search of the matrix m rooted at the vertex n.*/

pathlist(m) = {
  local(l=length(m),nexto=2,check=1,paths=vector(l), found=vector(l));
  paths[1] = [1];
  found[1] = 1;
  while(found[l]==0,
    for(i=1, l, 
      if(!aneq(anrat(0), m[found[check],i])&&!iselement(i, found), 
        found[nexto]=i;
        paths[nexto]=concat(paths[check],i);
        nexto++;
      if(found[l]!=0, break)));
    check++);
  paths[1]=[1,1];
  paths
}
addhelp(pathlist, "pathlist(m,n): a list of paths in the matrix m rooted at vertex n, found using a breadth-first search.")

/* This is the intermediate field defined as "little k (P)" in Maclachlan and
Reid section 10.4 */
littlek(g) =
{
  local(m=0, zero=anrat(0), z=anrat(1),
        n = matsize(g)[1],
        p = vector(n),
        f = vector(n));

  /* adjoin the squares of the elements of the Gram matrix */
  for(i=1,n,
    for(j=1,n,
      /*z = ancompositum(z, anmul(g[i,j], g[i,j]))));*/
      z = reducefield(ancompositum(z, anmul(g[i,j], g[i,j])))));
  
  /* adjoin products of elements along cycles */
  visit(1, -1);
  reducefield(z);
}

/* Depth first search done in littlek, uses "p" a global variable */
/*
otherpathlist(g) =
{
  local(n = matsize(g)[1], r = vector(n), c, j);
  for(i=1,n,
    c = listcreate(n);
    j = i; while(j!=1, listinsert(c, j, 1); j = p[j]);
    listinsert(c, 1, 1);
    r[i] = Vec(c));
  r
}
*/

visit(i, j) =
{
  local(a,b,y);

  p[i] = j; f[i] = m++;
  for(k=1,n,
    if(!aneq(g[i,k],zero) && k!=j,
      if(p[k],
        if(f[k]<f[i],
	   /* TO DO: rewrite using Greg's pathprod */
	   a = k; b = i; y = g[a,b];
	   while(b!=k, a=b; b=p[b]; y=anmul(y,g[a,b]));
	   z = reducefield(ancompositum(z, y))),
	   /*z = ancompositum(z, y)),*/
	visit(k,i))))
}
