/* graph theoretic functions */

spanning_tree(adj, root=1) =
{
  local(n = matsize(adj)[1],
        parent = vector(n),
        discovered = vector(n),
        pending = vector(n),
        front=0, back=0,
        v);
        
  pending[back++] = root;
  discovered[root] = 1;
  while(front < back,
    v = pending[front++];
    for(w=1, n,
      if(!discovered[w] && adj[v,w],
        pending[back++] = w;
        discovered[w] = 1;
        parent[w] = v)));
  return(parent)
}

path_from_root(tree, vertex) =
{
  local(path = listcreate(#tree));
  while(vertex!=0,
    listinsert(path, vertex, 1);
    vertex = tree[vertex]);
  return(Vec(path))
}

cycle_basis(adj, tree=0) =
{
  local(n = matsize(adj)[1],
        cycles = listcreate(binomial(n,2)-n+1),
        u_path, v_path, i);
        
  if(tree==0, tree = spanning_tree(adj));
  
  for(u=1, n,
    for(v=u+1, n,
      if(adj[u,v] && tree[u]!=v && tree[v]!=u,
        /* each non-tree edge creates a cycle */
        u_path = path_from_root(tree, u);
        v_path = path_from_root(tree, v);
        i=1; while(u_path[i] == v_path[i], i++); i--;
        listput(cycles,
          concat(vecextract(u_path, Str(i, "..", #u_path)),
                 vecextract(v_path, Str(#v_path, "..", i)))))));
  return(Vec(cycles))
}

adjacency_from_gram(G) =
{
  local(n = matsize(G)[1], zero=anrat(0));
  return(matrix(n,n,i,j,!aneq(G[i,j], zero)))
}

path_prod(G, p) = anprod(vector(#p-1, i, G[p[i],p[i+1]]))

Mmatrix(G, tree=0) =
{
  local(n = matsize(G)[1], zero = anrat(0), path_prods, M);
  
  if(tree==0, tree=spanning_tree(adjacency_from_gram(G)));
  path_prods = vector(4, i, path_prod(G, path_from_root(tree, i)));

  for(u=4, n,
    path_prods[4] = path_prod(G, path_from_root(tree, u));
    M = anmatinit(matrix(4,4,i,j,
          anprod([G[if(i==4,u,i),if(j==4,u,j)], path_prods[i], path_prods[j]])));
    if(!aneq(anmatdet(M),zero),
      if(verbose, print("used vertex ", u));
      return(M)));
  error("path products of G should be independent")
}

flatten(M) =
{
  local(s = matsize(M), m = s[1], n = s[2]);
  return(vector(m*n, i, M[(i-1)\n+1, (i-1)%n+1]))
}

extend_field(u, v) =
{
  if(type(v)=="t_MAT", v = flatten(v));
  for(i=1, #v, if(verbose,print("Compositing element ",i," of ",#v," elements"));u = reducefield(ancompositum(u, v[i])));
  return(u)
}

invariant_trace_field(G) = compute_field(G,0);

littlek(G) = compute_field(G,1);

both_fields(G) = compute_field(G,2);


compute_field(G,which) =
{
  local(n = matsize(G)[1],
        G_sq = matrix(n,n, i,j, ansubst(x^2,x,G[i,j])),
        squares = extend_field(anrat(0), G_sq),
        
        A = adjacency_from_gram(G),
        tree = spanning_tree(A),
        cycles = cycle_basis(A, tree),
        /*cycle_prods = vector(#cycles, i,path_prod(G, cycles[i])),*/
	cycle_prods,
	k,
        /*k = extend_field(squares, cycle_prods),*/
        itf, 
        M,
	disc_sqrt);
 
  if(verbose,print("Found ",matsize(cycles)," cycles")); 
  if(debug,print(cycles));
  cycle_prods = vector(#cycles, i,if(verbose,print("Computing cyclic product ",i));(path_prod(G, cycles[i])));
  k = extend_field(squares, cycle_prods);
  if(which==1,return(k));
  if(verbose,print("Computing the discriminant"));
  M = Mmatrix(G, tree);
  if(verbose,print("Little k is generated by: ",k));
  disc_sqrt =  ansqrt(anmatdet(M));
  if(debug,print("Pre-reduced square root of discriminant: ",disc_sqrt));
  disc_sqrt = reducefield(disc_sqrt);
  if(verbose,print("compositing the square root of the discriminant: ",disc_sqrt));
  itf =  extend_field(k, [disc_sqrt]);
  if(which == 0,return(itf),return([itf,k]));
}

